
#include <math.h>
#include <string.h>

#include "image.h"
#include "win-gl.h"

Image::Image (void)
{
	m_options = IMG_UNDEF;
	m_xres = 0; m_yres = 0;
	m_xsub = -1; m_ysub = -1;
	m_xpiv = 0.0; m_ypiv = 0.0;
	m_data = NULL; m_alpha = NULL; m_depth = NULL;	
	m_status = IMG_NOT_READY;
}

Image::Image (int xr, int yr, ImageOps ops)
{
	Size (xr, yr, ops);	
}

Image::~Image (void)
{
	if (m_data!=NULL) {delete [] m_data; m_data = NULL;}
	if (m_alpha!=NULL) {delete [] m_alpha; m_alpha = NULL;}
	if (m_depth!=NULL) {delete [] m_depth; m_depth = NULL;}
	m_xres = 0; m_yres = 0;
	m_xsub = 0; m_ysub = 0;
	m_options = IMG_UNDEF;	
}

void Image::Size (int xr, int yr, ImageOps ops)
{
	int bpp;
	
	if (m_data!=NULL) {delete [] m_data; m_data = NULL;}	// Destroy any previous main data
	if (m_alpha!=NULL) {delete [] m_alpha; m_alpha = NULL;}	// Destroy any previous alpha data
	if (m_depth!=NULL) {delete [] m_depth; m_depth = NULL;}	// Destroy any previous depth data

	m_options = ops;							// Set new pixel format
	bpp = GetBitsPerPixel ();					// Get bits per pixel
	m_size = (bpp*xr*yr)>>3;					// Calculate primary data size	
	m_data = new XBYTE[m_size];					// Allocate new primary data
	if (HasAlphaSeparate()) {					// Check if we need seperate alpha data..		
		m_alpha = new XBYTE[xr*yr];				// Yes. Allocate seperate alpha.	
	}
	if (HasDepthSeparate()) {					// Check if we need seperate depth data..
		m_depth = new float[xr*yr];				// Yes. Allocate seperate depth.
	}	
	m_xres = xr; m_yres = yr;	
	m_xsub = xr; m_ysub = yr;
}

inline unsigned char Image::GetBitsPerPixel (void)
{													// Pixel Format
	unsigned char bpp = 0;
	if (m_options & IMG_BW) bpp = 1;				// BPP = 1
	if (m_options & IMG_COLOR_INDEX) bpp = 8;		// BPP = 8
	if (m_options & IMG_COLOR_LOW) bpp = 12;		// BPP = 4-4-4
	if (m_options & IMG_COLOR_HIGH) bpp = 16;		// BPP = 6-5-5
	if (m_options & IMG_COLOR_TRUE) bpp = 24;		// BPP = 8-8-8	
	if (m_options & IMG_ALPHA_MERG) bpp += 8;		// BPP = x-x-x-8		
	if (m_options & IMG_DEPTH_MERG) bpp += 8;		// BPP = x-x-x-x-8
	return bpp;	
}

inline bool Image::HasAlpha (void)
{
	if (m_options & IMG_ALPHA_MERG) return true;
	if (m_options & IMG_ALPHA_SEP) return true;
	return false;
}

inline bool Image::HasAlphaMerged (void)
{
	if (m_options & IMG_ALPHA_MERG) return true;
	return false;
}

inline bool Image::HasAlphaSeparate (void)
{
	if (m_options & IMG_ALPHA_SEP) return true;
	return false;
}

inline bool Image::HasDepthMerged (void)
{
	if (m_options & IMG_DEPTH_MERG) return true;
	return false;
}

inline bool Image::HasDepth (void)
{
	if (m_options & IMG_DEPTH_MERG) return true;
	if (m_options & IMG_DEPTH_SEP) return true;
	return false;
}

inline bool Image::HasDepthSeparate (void)
{
	if (m_options & IMG_DEPTH_SEP) return true;
	return false;
}

void Image::SetPixel (int x, int y, XBYTE r, XBYTE g, XBYTE b)
{
	// ASSUMES 24-bit (RGB)
	if (x>=0 && y>=0 && x<m_xres && y<m_yres) {
		if (HasAlphaMerged()) {			
		} else {
			int offs = (y*m_xres+x)*3;
			*(m_data + offs) = r;
			*(m_data + offs+1) = g;
			*(m_data + offs+2) = b;
		}
	}
}

void Image::SetPixel (int x, int y, XBYTE r, XBYTE g, XBYTE b, XBYTE a)
{
	// ASSUMES 32-bit (RGB)
	if (x>=0 && y>=0 && x<m_xres && y<m_yres) {
		if (HasAlphaMerged()) {					
			XBYTE4 clr = r;
			clr += g<<8;
			clr += b<<16;
			clr += a<<24;
			*( ((XBYTE4*) m_data) + y*m_xres+x) = clr;
		}
	}
}


void Image::Fill (XBYTE r, XBYTE g, XBYTE b)
{
	if (HasAlphaMerged()) {

	} else {
		XBYTE fillbuf[6000];
		XBYTE *dest_pix, *dest_pix_stop;
		int xs = m_xres * 3;		
		for (int x=0; x < xs;) {fillbuf[x++] = r; fillbuf[x++] = g; fillbuf[x++] = b;}
		dest_pix = m_data;
		dest_pix_stop = m_data + m_yres * xs;
		for (; dest_pix < dest_pix_stop;) {
			memcpy (dest_pix, fillbuf, xs);
			dest_pix += xs;
		}
	}
}

bool Image::HasPowerSize (void)
{
	int x, y;
	GetPowerSize (x, y);
	if (x==m_xres && y==m_yres) return true;
	return false;
}

void Image::GetPowerSize (int &x, int &y)
{	
	x = GetSmallestPower (m_xres);
	y = GetSmallestPower (m_yres);
}

void Image::CopyPixels (int xoff, int yoff, Image& dest)
{
	TransferPixels (m_data, m_xres, m_yres, m_options, dest.m_data, dest.m_xres, dest.m_yres, dest.m_options, xoff, yoff);
}

void Image::ConvertToPowerImage (void)
{	
	if (!HasPowerSize()) {
		XBYTE* new_data;
		int new_x, new_y, new_size, bpp;
		GetPowerSize (new_x, new_y);
		bpp = GetBitsPerPixel ();					// Get bits per pixel
		new_size = (bpp*new_x*new_y)>>3;			// Calculate primary data size	
		new_data = new XBYTE[new_size];				// Allocate new primary data
		TransferPixels (m_data, m_xres, m_yres, m_options, new_data, new_x, new_y, m_options, 0, 0);
		delete [] m_data;
		m_xsub = m_xres;	m_ysub = m_yres;		// Save sub-image dimensions
		m_data = new_data;	m_size = new_size;		// Switch to new image
		m_xres = new_x;		m_yres = new_y;			// Switch to new dimensions
	}
}

/*inline int ImageX::GetFormat (ImageOps ops)
{
	if (ops & IMG_INDEXED) return IMG_INDEXED;
	if (ops & IMG_LOWCOLOR) return IMG_LOWCOLOR;
	if (ops & IMG_HIGHCOLOR) return IMG_HIGHCOLOR;
	if (ops & IMG_TRUECOLOR) return IMG_TRUECOLOR;
	return IMG_UNDEF;
}

inline int ImageX::GetAlpha (ImageOps ops)
{	
	if (ops & IMG_ALPHA) return IMG_ALPHA;
	return IMG_UNDEF;
}

inline void ImageX::TurnOff (ImageOps opt)
{	
	m_options ^= opt;
}

inline void ImageX::TurnOn (ImageOps opt)
{	
	m_options |= opt;
}


void ImageX::UseWith (int what)
{
	m_usage = what;
	ConvertFormat (IMG_HIGHCOLOR);
	switch (what) {
	case DRAW_IMAGE:					break;
	case DRAW_MASKED:					ConvertAlphaToMask(); break;
	case DRAW_ALPHA: case DRAW_ADDED:	ConvertAlphaToBlend(); break;
	}
	status = IMG_OK;
}

void ImageX::ConvertFormat (ImageOps ops)
{	
	int n;
	int v;
	XBYTE * newdata;
	
	switch (GetFormat(m_options)) {			// Source Format
	case IMG_TRUECOLOR: {					//   True Color (24 bit)
		switch (GetFormat(ops))	{			// Target Format
		case IMG_HIGHCOLOR: {				//   High Color (16 bit)
			XBYTE *pDataOld;
			XBYTE2 *pDataNew;
			newdata = new XBYTE[m_xres * m_yres * GetBytesPerPixel(ops)];			
			pDataOld = (XBYTE *) m_data; 
			pDataNew = (XBYTE2 *) newdata;
			for (n=0; n<m_xres*m_yres; n++) {
				v = (*pDataOld++ << win_maxr) >> 8; *pDataNew = v << win_shiftr;
				v = (*pDataOld++ << win_maxg) >> 8;	*pDataNew |= v << win_shiftg;
				v = (*pDataOld++ << win_maxb) >> 8; *pDataNew++ |= v << win_shiftb;
			}
			delete [] m_data;
			m_data = newdata;			
		} break;
		}
	} break;
	}
	TurnOff (GetFormat(m_options));
	TurnOn (GetFormat(ops));
}

void ImageX::ConvertAlphaToMask (void)
{
	// NOTE: IF NO ALPHA CHANNEL IS LOADED, any pixels with value 0 
	// in the primary data with be treated as transparent. DrawMasked
	// will draw all 0 value pixels transparent. This is the desired behavior.
	// If an alpha channel >is< present, 0 value pixels will be mapped to value 1,
	// and the alpha channel will be used to determine which pixels to make transparent
	// by setting them to 0.
	if (GetAlpha(m_options)==IMG_ALPHA) {
		switch (GetFormat(m_options)) {
		case IMG_HIGHCOLOR: {
			XBYTE2 *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_xres*m_yres; n++) {
				if (*pData==0) *pData = 1;
				if (*pAlphaData==0) *pData = 0;
				*pData++;
				*pAlphaData++;
			}
		} break;
		}
		delete [] m_alpha;
		m_alpha = NULL;
		TurnOff (IMG_ALPHA);
		TurnOn (IMG_MASKED);
	}
}

void ImageX::ConvertAlphaToBlend (void)
{
	if (GetAlpha(m_options)==IMG_ALPHA) {		
		switch (GetFormat(m_options)) {
		case IMG_TRUECOLOR: {			
			XBYTE *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_xres*m_yres; n++) {
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pAlphaData++ = 255 - *pAlphaData;
			}
		} break;
		case IMG_HIGHCOLOR: {
			XBYTE2 r, g, b;
			XBYTE2 *pData;
			XBYTE *pAlphaData;			
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;			
			for (int n=0; n<m_xres*m_yres; n++) {				
				r = (((*pData & win_maskr) >> win_shiftr) * *pAlphaData) >> 8;
				g = (((*pData & win_maskg) >> win_shiftg) * *pAlphaData) >> 8;
				b = (((*pData & win_maskb) >> win_shiftb) * *pAlphaData) >> 8;				
				*pData++ = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
				*pAlphaData = 255 - *pAlphaData;
				pAlphaData++;
			}
		} break;
		}		
	} else {										// NO ALPHA CHANNEL LOADED
		if (m_alpha!=NULL) delete [] m_alpha;	
		m_alpha = new XBYTE[m_xres * m_yres];		// Create an alpha channel	
		TurnOn (IMG_ALPHA);		
		switch (GetFormat(m_options)) {
		case IMG_TRUECOLOR: {			
			XBYTE *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_xres*m_yres; n++) {	// Alpha channel is gray of primary data
				*pAlphaData = (*pData + *(pData+1) + *(pData+2))/3;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pAlphaData++ = 255 - *pAlphaData;
			}
		} break;
		case IMG_HIGHCOLOR: {
			XBYTE2 r, g, b;
			XBYTE2 *pData;
			XBYTE *pAlphaData;			
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;			
			for (int n=0; n<m_xres*m_yres; n++) {				
				r = (*pData & win_maskr) >> win_shiftr;
				g = (*pData & win_maskg) >> win_shiftg;
				b = (*pData & win_maskb) >> win_shiftb;
				*pAlphaData = ((r + g + b)*4)/3;
				r = (r * *pAlphaData) >> 8;
				g = (g * *pAlphaData) >> 8;
				b = (b * *pAlphaData) >> 8;
				*pData++ = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
				*pAlphaData = 255 - *pAlphaData;
				pAlphaData++;
			}
		} break;
		}
	}
}

int ImageX::GetIntensity (int x, int y)
{
	int r, g, b;
	XBYTE2 *pData = (XBYTE2*) m_data;
	pData += y*m_xres + x;
	r = (*pData & win_maskr) >> win_shiftr;
	g = (*pData & win_maskg) >> win_shiftg;
	b = (*pData & win_maskb) >> win_shiftb;
	return ((r + g + b)*4)/3;
}


void ImageX::Rotate (ImageX *dest, float ang, float scale)
{	
	dest->SetUsage (m_usage);
	RotateImage (dest, ang, scale);
	RotateAlpha (dest, ang, scale);
}
		
void ImageX::RotateImage (ImageX *dest, float ang, float scale)
{
	// NOTE: ASSUMES BOTH IMAGES ARE IMG_HIGHCOLOR

	XBYTE2 *src_pix;
	XBYTE2 *dest_pix;
	XBYTE2 *dest_start;
	float delt_x, delt_y;
	float start_x, start_y;
	float off_x, off_y;
	float pix_x, pix_y, h;
	int y, cnt;

	if (dest!=NULL && m_data!=NULL) {
	
	dest->SetUsage(m_usage);
	h = (float) sqrt(dest->m_xres * dest->m_xres + dest->m_yres * dest->m_yres);
	start_x = (m_xres/2.0f) + (float) cos((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
	start_y = (m_yres/2.0f) - (float) sin((135-ang)*DEGtoRAD)*(h/2.0f)*(m_xres/m_yres)/scale;
	delt_x = (float) cos(ang*DEGtoRAD)/scale;
	delt_y = (float) sin(ang*DEGtoRAD)*(m_xres/m_yres)/scale;
	off_x = delt_x * dest->m_xres;
	off_y = delt_y * dest->m_xres;	
	dest_start = (XBYTE2*) dest->m_data;	
	for (y=0; y<dest->m_yres; y++) {	
		cnt = dest->m_xres;						// Start with full pixel set in line
		pix_x = start_x + off_x;				// Goto end of line and scan backwards
		pix_y = start_y + off_y;
		dest_pix = dest_start + cnt-1;
		while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
			pix_x -= delt_x;					// Scan until end of span found
			pix_y -= delt_y;
			cnt--;
			*dest_pix-- = 0;
		}	
		dest_pix = dest_start;
		pix_x = start_x;						// Goto beginning of line and scan forward
		pix_y = start_y;
		while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
			pix_x += delt_x;					// Scan until start of span found
			pix_y += delt_y;
			cnt--;
			*dest_pix++ = 0;
		}
		while (cnt>0) {
			src_pix = ((XBYTE2*) m_data) + (int) pix_y*m_xres + (int) pix_x;
			*dest_pix++ = *src_pix;			
			pix_x += delt_x; pix_y += delt_y;
			cnt--;
		}
		dest_start += dest->m_xres;
		start_x -= delt_y;
		start_y += delt_x;
	}

	}
}

void ImageX::RotateAlpha (ImageX *dest, float ang, float scale)
{
	XBYTE *src_pix;
	XBYTE *dest_pix;
	XBYTE *dest_start;
	float delt_x, delt_y;
	float start_x, start_y;
	float off_x, off_y;
	float pix_x, pix_y, h;
	int y, cnt;

	if (m_alpha!=NULL && dest->m_alpha!=NULL) {

		dest->SetUsage(m_usage);
		h = (float) sqrt(dest->m_xres * dest->m_xres + dest->m_yres * dest->m_yres);
		start_x = (m_xres/2.0f) + (float) cos((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		start_y = (m_yres/2.0f) - (float) sin((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		delt_x = (float) cos(ang*DEGtoRAD)/scale;
		delt_y = (float) sin(ang*DEGtoRAD)/scale;
		off_x = delt_x * dest->m_xres;
		off_y = delt_y * dest->m_xres;	
		dest_start = (XBYTE*) dest->m_alpha;
		for (y=0; y<dest->m_yres; y++) {	
			cnt = dest->m_xres;						// Start with full pixel set in line
			pix_x = start_x + off_x;				// Goto end of line and scan backwards
			pix_y = start_y + off_y;
			dest_pix = dest_start + cnt-1;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x -= delt_x;					// Scan until end of span found
				pix_y -= delt_y;
				cnt--;
				*dest_pix-- = 255;
			}	
			dest_pix = dest_start;
			pix_x = start_x;						// Goto beginning of line and scan forward
			pix_y = start_y;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x += delt_x;					// Scan until start of span found
				pix_y += delt_y;
				cnt--;
				*dest_pix++ = 255;
			}
			while (cnt>0) {
				src_pix = ((XBYTE*) m_alpha) + (int) pix_y*m_xres + (int) pix_x;
				*dest_pix++ = *src_pix;			
				pix_x += delt_x; pix_y += delt_y;
				cnt--;
			}
			dest_start += dest->m_xres;
			start_x -= delt_y;
			start_y += delt_x;
		}
	}
}


ImageExt::ImageExt ()
{
	m_pImg = NULL;
}

ImageExt::ImageExt (ImageX* pImg)
{
	m_pImg = pImg;
}

ImageExt::~ImageExt ()
{
	m_pImg = NULL;
}

void ImageExt::Load (char *filename, ImageX *img, int what)
{
	LoadTiff (filename, img);
	img->UseWith (what);	
}

*/
static int GetSmallestPower (int x)
{
	int i = 1;
	for (; i < x; i <<= 1);
	return i;
}

// Clip Pixels
//  
// Input:
//		src_wid		Width of source pixel buffer
//		src_hgt		Height of source pixel buffer
//		dest_wid	Width of desination pixel buffer
//		dest_hgt	Height of desination pixel buffer
//		xoff,yoff	Position in desination to place source 
//		vx1, vy1	Top-Left clipping region in destination
//		vx2, vy2	Bottom-Right clipping region in destination
// Output:
//		src_offs	Starting offset in source buffer (in pixels)
//		src_xs		Number of pixels to draw horizontally
//		src_ys		Number of pixels to draw vertically
//		dest_offs	Starting offset in destination buffer (in pixels)
//
static bool ClipPixels (int src_wid, int src_hgt, int dest_wid, int dest_hgt, int xoff, int yoff, int vx1, int vy1, int vx2, int vy2, int &src_offs, int &src_xs, int &src_ys, int &dest_offs)
{	
	int xend = xoff + (src_wid-1);
	int yend = yoff + (src_hgt-1);
	src_xs = src_wid;
	src_ys = src_hgt;

	if (xend >= vx1 && xoff <= vx2) {					// In-View Horizontally
		if (yend >= vy1 && yoff <= vy2) {				// In-View Vertically 
			dest_offs = vy1*dest_wid + vx1;			// Assume no clipping
			src_offs = 0;								
			if (xoff < vx1) {							// Clip Left
				src_xs -= vx1 - xoff;					// (Reduce number of horiz. pixels drawn)
				src_offs += vx1 - xoff;					// (Move into src buffer)
			} else {
				dest_offs += xoff - vx1;				// (Position image in dest buffer)
			}
			if (xend > vx2)								// Clip Right
				src_xs -= xend - vx2;
			
			if (yoff < vy1) {							// Clip Top
				src_ys -= vy1 - yoff;
				src_offs += (vy1 - yoff) * src_wid;
			} else {
				dest_offs += (yoff - vy1) * dest_wid;
			}
			if (yend > vy2)								// Clip Bottom
				src_ys -= yend - vy2;
			return true;
		}
	}
	return false;
}

static bool ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy)
{
	if (x<0) {
		mx = ((int) (x / xres)) - 1;
		offx = (x % xres) + xres;
	} else {
		mx = ((int) (x / xres));
		offx = x % xres;
	}
	if (y<0) {
		my = ((int) (y / yres)) - 1;		
		offy = (y % yres) + yres;
	} else {
		my = ((int) (y / yres));		
		offy = y % yres;
	}	
}



static bool TransferPixels (XBYTE* src, int src_wid, int src_hgt, ImageOps src_ops, XBYTE* dest, int dest_wid, int dest_hgt, ImageOps dest_ops, int xoff, int yoff)
{
	int src_offs, dest_offs;
	int src_xs, src_ys;	

	// THIS FUNCTION ONLY WORKS IF SRC AND DEST ARE BOTH 24-bit COLOR FORMAT
	// (However, they may or may not contain an integrated alpha channel)
	if ((src_ops & IMG_COLOR_TRUE) & (dest_ops & IMG_COLOR_TRUE)) {
		if (src_ops & IMG_ALPHA_MERG) {
			if (dest_ops & IMG_ALPHA_MERG) {				
				register XBYTE4 *src_pix, *src_pix_stop, *dest_pix;
				if (ClipPixels (src_wid, src_hgt, dest_wid, dest_hgt, xoff, yoff, 0, 0, dest_wid-1, dest_hgt-1, src_offs, src_xs, src_ys, dest_offs)) {
					src_xs <<= 2;
					dest_pix = ((XBYTE4*) dest) + dest_offs;
					src_pix = ((XBYTE4*) src) + src_offs;					
					src_pix_stop = src_pix + (src_ys * src_wid);
					for (; src_pix < src_pix_stop;) {
						memcpy (dest_pix, src_pix, src_xs);
						dest_pix += dest_wid;
						src_pix += src_wid;
					}
				}
				return true;
			} else {
				return false;
			}
		} else {
			if (dest_ops & IMG_ALPHA_MERG) {
				return false;
			} else {
				register XBYTE *src_pix, *src_pix_stop, *dest_pix;
				if (ClipPixels (src_wid, src_hgt, dest_wid, dest_hgt, xoff, yoff, 0, 0, dest_wid-1, dest_hgt-1, src_offs, src_xs, src_ys, dest_offs)) {
					src_xs *= 3;
					dest_pix = dest + (dest_offs*3);
					src_pix = src + (src_offs*3);
					src_pix_stop = src_pix + (src_ys * src_wid*3);
					dest_wid *= 3; src_wid *= 3;
					for (; src_pix < src_pix_stop;) {
						memcpy (dest_pix, src_pix, src_xs);
						dest_pix += dest_wid;
						src_pix += src_wid;
					}
				}
				return true;
			}
		}
	} else {
		return false;
	}
	return false;
}

int Image::GetGLFormat (void)
{
	if (m_options & IMG_ALPHA_MERG) {
		return GL_RGBA;
	} else {
		return GL_RGB;
	}
}