//
// GameX
// Ship Demo - Photon Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
#include <math.h>

#include "win.h"
#include "shipnet-eng.h"
#include "shipnet-particles.h"

void Particles::Initialize (void)
{
	char tiffile[100];
	ImageExt imgext;

	max_particles = 2000;
	particle_list = new Particle[max_particles];
	
	imgext.Load ("Images\\photon1.tif", &photon_img[0]);
	imgext.Load ("Images\\photon2.tif", &photon_img[1]);
	imgext.Load ("Images\\photon3.tif", &photon_img[2]);
	photon_img[0].SetPivot (25, 25);

	for (int n=0; n < 14; n++) {		
		sprintf (tiffile, "Images\\explode%02d.tif", n+1);
		imgext.Load (tiffile, &explosion_img[n]);
		explosion_img[n].SetPivot (128, 128);
	}
}

// Particles - AddPhoton
// Add a photon to the list of particles.
int Particles::AddPhoton (char ship, float new_x, float new_y, float new_dx, float new_dy)
{
	int new_particle;
	if (num_particles==max_particles) {		// If we've reached maximum, select an existing
										// particle at random to replace.
		new_particle = (rand() * (max_particles-1))/RAND_MAX;
	} else {							
		new_particle = num_particles;		// If there is still allocated memory left,
		num_particles++;					// add new particle at the end
	}
	particle_list[new_particle].src_ship = ship;
	particle_list[new_particle].kind = P_PHOTON;
	particle_list[new_particle].x = new_x;	// Initialize new photon with given parameters
	particle_list[new_particle].y = new_y;
	particle_list[new_particle].dx = new_dx;
	particle_list[new_particle].dy = new_dy;
	particle_list[new_particle].phase = (rand()*360)/RAND_MAX;
	particle_list[new_particle].age = 0;	
	particle_list[new_particle].img = 0;
	return new_particle;					// Return the number of the photon just added
}

int Particles::AddExplosion (char ship, float new_x, float new_y, float new_dx, float new_dy)
{
	int new_particle;
	if (num_particles==max_particles) {		// If we've reached maximum, select an existing
										// particle at random to replace.
		new_particle = (rand() * (max_particles-1))/RAND_MAX;
	} else {							
		new_particle = num_particles;		// If there is still allocated memory left,
		num_particles++;					// add new particle at the end
	}
	particle_list[new_particle].src_ship = ship;
	particle_list[new_particle].kind = P_EXPLOSION;
	particle_list[new_particle].x = new_x;	// Initialize new photon with given parameters
	particle_list[new_particle].y = new_y;
	particle_list[new_particle].dx = new_dx;
	particle_list[new_particle].dy = new_dy;
	particle_list[new_particle].phase = (rand()*30)/RAND_MAX;
	particle_list[new_particle].age = -1;
	particle_list[new_particle].img = 0;
	return new_particle;					// Return the number of the photon just added
}

// Particles - Move
// Move all the Particles in the particle system.
// The Particles are implemented as a very simple physics engine for 
// a particle system. Each particle is advanced according to its velocity.
// Particles out of bounds are rebounded into the view.
// Particles which are too old are deleted.
void Particles::Move (void)
{ 
	int max_age;
										// Process all particles (Particles) for moving
	for (int n=0; n<num_particles; n++) {
		particle_list[n].x += particle_list[n].dx;
		particle_list[n].y += particle_list[n].dy;		
		switch (particle_list[n].kind) {
		case P_PHOTON: {
			max_age = 400;
			particle_list[n].age++;
		} break;
		case P_REMNANT: {
			particle_list[n].dx *= 0.9;
			particle_list[n].dy *= 0.9;
			particle_list[n].age++;
			max_age = 60;
		} break;
		case P_EXPLOSION: {
			if (particle_list[n].age==-1) {
				if (--particle_list[n].phase <= 0.0) {
					particle_list[n].age = 0;
					particle_list[n].phase = (int) (rand()*360.0)/RAND_MAX;
				}
			} else {
				particle_list[n].age++;				
			}						
			max_age = 14;
		} break;
		}
		if (particle_list[n].age > max_age) {			// If particle is too old..
			particle_list[n].src_ship = particle_list[num_particles-1].src_ship;
			particle_list[n].kind = particle_list[num_particles-1].kind;
			particle_list[n].x = particle_list[num_particles-1].x;	// Delete it putting the
			particle_list[n].y = particle_list[num_particles-1].y;	// last one in its place.
			particle_list[n].dx = particle_list[num_particles-1].dx;
			particle_list[n].dy = particle_list[num_particles-1].dy;
			particle_list[n].age = particle_list[num_particles-1].age;
			particle_list[n].phase = particle_list[num_particles-1].phase;
			num_particles--;										// Decrement total number.
		}		
	}
}

int Particles::DetectCollisions (char chk_ship, float sx, float sy, float sdx, float sdy, float *ang)
{
	float deltx, delty, dist, speedj;
	float recoilx, recoily, dot;
	int cp = -1;
	for (int n=0; n < num_particles; n++) {
		if (particle_list[n].kind==P_PHOTON && particle_list[n].src_ship != chk_ship) {
			deltx = sx - particle_list[n].x;
			delty = sy - particle_list[n].y;
			dist = sqrt (deltx*deltx + delty*delty);

			if (dist < SHIELD_SIZE+PARTICLE_SIZE) {
				particle_list[n].x -= particle_list[n].dx;
				particle_list[n].y -= particle_list[n].dy;
				deltx = particle_list[n].x - sx;
				delty = particle_list[n].y - sy;
				dist = sqrt (deltx*deltx + delty*delty);				
				speedj = sqrt (particle_list[n].dx*particle_list[n].dx + particle_list[n].dy*particle_list[n].dy);
				
				deltx /= dist;
				delty /= dist;

				*ang = (atan2 (delty, deltx)/DEGtoRAD)+85.0 + (rand()*10.0f/RAND_MAX);

				dot = particle_list[n].dx * -delty + particle_list[n].dy * deltx;
				recoilx = delty*dot*2.0f;
				recoily = -deltx*dot*2.0f;
				particle_list[n].kind = P_REMNANT;
				particle_list[n].age = 0;
   				particle_list[n].dx += recoilx;
				particle_list[n].dy += recoily;				
				float damp = 0.2 + rand()*0.7/RAND_MAX;
				particle_list[n].dx = sdx-particle_list[n].dx*damp;
				particle_list[n].dy = sdy-particle_list[n].dy*damp;
				particle_list[n].x += particle_list[n].dx;
				particle_list[n].y += particle_list[n].dy;
				
				/*particle_list[n].src_ship = particle_list[num_particles-1].src_ship;
				particle_list[n].kind = particle_list[num_particles-1].kind;
				particle_list[n].x = particle_list[num_particles-1].x;	// Delete it putting the
				particle_list[n].y = particle_list[num_particles-1].y;	// last one in its place.
				particle_list[n].dx = particle_list[num_particles-1].dx;
				particle_list[n].dy = particle_list[num_particles-1].dy;
				particle_list[n].age = particle_list[num_particles-1].age;
				particle_list[n].phase = particle_list[num_particles-1].phase;
				num_particles--;										*/
				cp = n;
			}
		}
	}
	return cp;
}
	

// Particles - Draw
// Draw all Particles
void Particles::Draw (float dx, float dy, int frame)
{
	float i, s;					// Process all particles (photons) for drawing
	int cx, cy;

	cx = WinGL.GetWinX()/2;
	cy = WinGL.GetWinY()/2;

	for (int n=0; n<num_particles; n++) {
		switch (particle_list[n].kind) {
		case P_PHOTON: {
			i = (rand()*360.0f)/RAND_MAX;		
			s = 0.7f + (float) sin((frame+particle_list[n].phase)*12*3.141592/180)*0.3f;
			WinGL.DrawImage (photon_img[0], int(particle_list[n].x-dx+cx), int(particle_list[n].y-dy+cy), i, s);
		} break;
		case P_REMNANT: {
			i = (rand()*360.0f)/RAND_MAX;		
			s = 0.7f + (float) sin((frame+particle_list[n].phase)*12*3.141592/180)*0.3f;
			glColor4f (0.9f, 0.9f, 0.9f, 1.0f);
			WinGL.DrawImage (photon_img[0], int(particle_list[n].x-dx+cx), int(particle_list[n].y-dy+cy), i, s);
			glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
		} break;
		case P_EXPLOSION: {
			if (particle_list[n].age >= 0) {
				s = 0.8f + float(particle_list[n].age*0.5f/14.0);
				WinGL.DrawImage (explosion_img[particle_list[n].age], int(particle_list[n].x-dx+cx), int(particle_list[n].y-dy+cy), (float) particle_list[n].phase, s);
			}
		} break;
		}
	}
}

